@@ -12,6 +12,12 @@ except ImportError:
     # Not available in Python 3.5
     FormattedValue = Constant = None
 
+try:
+    from ast import MatchAs, MatchSequence, MatchStar
+except ImportError:
+    # Not available before Python 3.10
+    MatchAs = MatchSequence = MatchStar = None
+
 CONTEXT_NAME = '__live_coding_context__'
 RESULT_NAME = '__live_coding_result__'
 
@@ -147,27 +153,7 @@ class Tracer(NodeTransformer):
         @return: format_text, next_index_to_get
         """
         slice_node = subscript_node.slice
-        if isinstance(slice_node, (Index, Ellipsis)):
-            if (isinstance(slice_node, Ellipsis) or
-                    isinstance(slice_node.value, Ellipsis)):
-                index_to_get = None
-                format_text = '...'
-            else:
-                slice_node.value = self._wrap_assignment_index(
-                    slice_node.value,
-                    index_to_get)
-                format_text = '{!r}'
-                if index_to_get is not None:
-                    index_to_get -= 1
-        elif Constant is not None and isinstance(slice_node, (Constant, Name)):
-            # Python 3.9 stopped wrapping the constant in an Index object.
-            subscript_node.slice = self._wrap_assignment_index(
-                slice_node,
-                index_to_get)
-            format_text = '{!r}'
-            if index_to_get is not None:
-                index_to_get -= 1
-        elif isinstance(slice_node, Slice):
+        if isinstance(slice_node, Slice):
             index_to_get = None
             if slice_node.step is None:
                 step_text = ''
@@ -197,8 +183,7 @@ class Tracer(NodeTransformer):
                           for element in slice_node.dims]
             format_text = ', '.join(self._wrap_slice(subscript)[0]
                                     for subscript in subscripts)
-        else:
-            assert isinstance(slice_node, Tuple), slice_node
+        elif isinstance(slice_node, Tuple):
             # Python 3.9 replaced ExtSlice with a Tuple.
             index_to_get = None
             subscripts = [Subscript(slice=element)
@@ -206,6 +191,28 @@ class Tracer(NodeTransformer):
             format_text = ', '.join(self._wrap_slice(subscript)[0]
                                     for subscript in subscripts)
             slice_node.elts = [subscript.slice for subscript in subscripts]
+        elif isinstance(slice_node, (Index, Ellipsis)):
+            if (isinstance(slice_node, Ellipsis) or
+                    isinstance(slice_node.value, Ellipsis)):
+                index_to_get = None
+                format_text = '...'
+            else:
+                slice_node.value = self._wrap_assignment_index(
+                    slice_node.value,
+                    index_to_get)
+                format_text = '{!r}'
+                if index_to_get is not None:
+                    index_to_get -= 1
+        else:
+            # Python 3.9 stopped wrapping many things in an Index object, so
+            # subscript_node.slice could be a Constant, BinOp, UnaryOp, or Call.
+            subscript_node.slice = self._wrap_assignment_index(
+                slice_node,
+                index_to_get)
+            format_text = '{!r}'
+            if index_to_get is not None:
+                index_to_get -= 1
+
         return format_text, index_to_get
 
     def visit_Call(self, node):
@@ -705,22 +712,31 @@ class Tracer(NodeTransformer):
         """
         if isinstance(target, Name):
             return target.id
+        if MatchAs is not None and isinstance(target, MatchAs):
+            return target.name
+        if MatchSequence is not None and isinstance(target, MatchSequence):
+            return self._wrap_assignment_sequence(target.patterns)
         if isinstance(target, Subscript):
             return self._wrap_subscript_target(target, index_to_get)
         if isinstance(target, Tuple) or isinstance(target, List):
-            target_names = map(self._wrap_assignment_target, target.elts)
-            wrapped = '({})'.format(', '.join(target_names))
-            if len(target.elts) == 1:
-                wrapped = wrapped[:-1] + ',)'
-            return wrapped
+            return self._wrap_assignment_sequence(target.elts)
         if Starred is not None and isinstance(target, Starred):
             return '*{}'.format(target.value.id)
+        if MatchStar is not None and isinstance(target, MatchStar):
+            return '*{}'.format(target.name)
         if not isinstance(target, Attribute):
             raise ValueError('Assignment target had type {}.'.format(
                 type(target)))
         names = self._get_attribute_names(target)
         return '.'.join(names)
 
+    def _wrap_assignment_sequence(self, targets):
+        target_names = map(self._wrap_assignment_target, targets)
+        wrapped = '({})'.format(', '.join(target_names))
+        if len(targets) == 1:
+            wrapped = wrapped[:-1] + ',)'
+        return wrapped
+
     def _wrap_assignment_targets(self, targets):
         """ Build string describing assignment targets and wrap indexes.
 
