@@ -18,15 +18,16 @@ try:
     # noinspection PyUnresolvedReferences
     from js import document, window
     IS_PYODIDE = True
-    MockTurtle = get_terminal_size = None
+    get_terminal_size = None
 except ImportError:
     IS_PYODIDE = False
     document = window = None
     from os import get_terminal_size
-    try:
-        from .mock_turtle import MockTurtle
-    except ImportError:
-        MockTurtle = None
+
+try:
+    from .mock_turtle import MockTurtle
+except ImportError:
+    MockTurtle = None
 
 from .canvas import Canvas
 from .code_tracer import CONTEXT_NAME, find_line_numbers
@@ -183,23 +184,31 @@ def main():
         exit(return_code)
 
 
-def analyze(source_code):
+def analyze(source_code, canvas_size=None):
     """ Trace the source code for display in the browser.
 
     :param source_code: Source code to trace.
+    :param canvas_size: (width, height), set if the report should include turtle
+        commands.
     :return: (tracing_report, output)
     """
     tracer = TraceRunner()
     tracer.standard_files.old_files['stderr'] = StringIO()
     tracer.max_width = 200000
     with replace_input(source_code):
-        code_report = tracer.trace_command(['space_tracer',
-                                            '--traced_file', PSEUDO_FILENAME,
-                                            '--source_width', '0',
-                                            '--live',
-                                            '--stdout', '!',
-                                            '--stderr', '!',
-                                            PSEUDO_FILENAME])
+        tracer_args = ['space_tracer',
+                       '--traced_file', PSEUDO_FILENAME,
+                       '--source_width', '0',
+                       '--live',
+                       '--stdout', '!',
+                       '--stderr', '!']
+        if canvas_size is not None:
+            canvas_width, canvas_height = canvas_size
+            tracer_args.append('--canvas')
+            tracer_args.append('-x{}'.format(canvas_width))
+            tracer_args.append('-y{}'.format(canvas_height))
+        tracer_args.append(PSEUDO_FILENAME)
+        code_report = tracer.trace_command(tracer_args)
     stdout = tracer.standard_files.old_files['stderr'].getvalue()
     return code_report, stdout
 
@@ -257,6 +266,9 @@ def traced(target=None, hide=None):
         # 3. module_importers.py that executed the traced module.
         this_filepath = Path(__file__)
         module_importers_filepath = this_filepath.parent / "module_importers.py"
+
+        if len(call_stack) < 4:
+            return False
         expected_frame = call_stack[3]
         return expected_frame.filename == str(module_importers_filepath)
 
@@ -317,6 +329,8 @@ def replace_input(stdin_text=None):
 
 
 def display_error_on_canvas():
+    if MockTurtle is None:
+        return
     tb = traceback.TracebackException(*sys.exc_info())
     tb_stack = tb.stack
     library_path = os.path.dirname(__file__)
@@ -364,17 +378,21 @@ class TraceRunner(object):
         self.return_code = None
         self.standard_files = StandardFiles()
 
-    def trace_turtle(self, source):
-        with replace_input(source):
-            self.trace_command(['space_tracer',
-                                '--traced_file', PSEUDO_FILENAME,
-                                '--source_width', '0',
-                                '--width', '0',
-                                '--height', '0',
-                                '--live',
-                                PSEUDO_FILENAME])
+    def trace_turtle(self, source, width: int = 0, height: int = 0):
+        MockTurtle.monkey_patch()
+        try:
+            with replace_input(source):
+                self.trace_command(['space_tracer',
+                                    '--traced_file', PSEUDO_FILENAME,
+                                    '--source_width', '0',
+                                    '--width', str(width),
+                                    '--height', str(height),
+                                    '--live',
+                                    PSEUDO_FILENAME])
 
-        return '\n'.join(MockTurtle.get_all_reports())
+            return '\n'.join(MockTurtle.get_all_reports())
+        finally:
+            MockTurtle.remove_monkey_patch()
 
     def trace_code(self, source):
         """ Trace a module of source code.
@@ -396,8 +414,12 @@ class TraceRunner(object):
         args = parse_args(command_args)
         if self.canvas is None:
             self.canvas = Canvas(args.width, args.height)
+        was_patched = False
         if MockTurtle is not None:
-            MockTurtle.monkey_patch(self.canvas)
+            if MockTurtle.is_patched():
+                was_patched = True
+            else:
+                MockTurtle.monkey_patch(self.canvas)
         self.standard_files['stdin'] = args.stdin
         self.standard_files['stdout'] = args.stdout
         self.standard_files['stderr'] = args.stderr
@@ -567,6 +589,8 @@ class TraceRunner(object):
             turtle_report = None
         else:
             turtle_report = MockTurtle.get_all_reports()
+            if not was_patched:
+                MockTurtle.remove_monkey_patch()
         if turtle_report and args.canvas:
             report = ('start_canvas\n' +
                       '\n'.join(turtle_report) +
