@@ -1,11 +1,9 @@
 from ast import parse
-from base64 import standard_b64encode
 import builtins
 from importlib.abc import MetaPathFinder, Loader
 from importlib.machinery import ModuleSpec
 from importlib.util import find_spec
 import inspect
-import io
 import os
 import sys
 import types
@@ -14,13 +12,16 @@ from pathlib import Path
 from .code_tracer import trace_source_tree, CONTEXT_NAME
 from .traced_finder import DEFAULT_MODULE_NAME, LIVE_MODULE_NAME, \
     PSEUDO_FILENAME, TracedFinder
+
 try:
-    from .mock_turtle import MockTurtle, monkey_patch_pyglet
+    from .mock_turtle import MockTurtle
+    from .live_image import LiveFigure, monkey_patch_pyglet
 except ImportError:
-    MockTurtle = monkey_patch_pyglet = None
+    MockTurtle = monkey_patch_pyglet = LiveFigure = None
 
 
 class DelegatingModuleFinder(MetaPathFinder):
+    # noinspection PyMethodOverriding
     def find_spec(self, fullname, path, target):
         for finder in self.following_finders:
             finder_find_spec = getattr(finder, 'find_spec', None)
@@ -395,8 +396,10 @@ class PatchedModuleLoader(Loader):
         if self.main_loader is not None:
             self.main_loader.exec_module(module)
         if self.fullname in ('numpy.random', 'random'):
+            # noinspection PyUnresolvedReferences
             module.seed(0)
         elif self.fullname == 'matplotlib':
+            # noinspection PyUnresolvedReferences
             module.use('Agg')
         elif self.fullname == 'matplotlib.pyplot':
             self.plt = module
@@ -429,14 +432,12 @@ class PatchedModuleLoader(Loader):
             y = (screen_height - figure_height) // 2
         else:
             y = 0
-        # noinspection PyUnresolvedReferences
-        data = io.BytesIO()
-        self.plt.savefig(data, format='PNG')
 
-        image = data.getvalue()
-        encoded = standard_b64encode(image)
-        image_text = str(encoded.decode('UTF-8'))
-        MockTurtle.display_image(x, y, image=image_text)
+        # Adjust to turtle coordinates.
+        x -= screen_width // 2
+        y = screen_height // 2 - y
+
+        LiveFigure(figure).display((x, y))
 
     def live_coding_zoom(self):
         screen_width, screen_height = self.plt.live_coding_size
