@@ -17,9 +17,6 @@ class CodeTracerTest(ReportTestCase):
         super(CodeTracerTest, self).setUp()
         self.maxDiff = None
 
-    def tearDown(self):
-        MockTurtle.remove_monkey_patch()
-
     def test_empty(self):
         # EXEC
         report = TraceRunner().trace_code("")
@@ -127,6 +124,8 @@ n -= 1
 
 
 IndentationError: expected an indented block """
+        if sys.version_info >= (3, 10):
+            expected_report += "after 'if' statement on line 3"
         tracer = TraceRunner()
 
         # EXEC
