@@ -1,16 +1,19 @@
-import re
-
 import pytest
 import sys
 
 from space_tracer.main import TraceRunner
-from test_report_builder import trim_report
+from test_live_image import replace_image
 
-MATPLOTLIB_UNSUPPORTED_VERSION = (3, 10)
+try:
+    import matplotlib
+except ImportError:
+    matplotlib = None
 
 
 @pytest.fixture(name='is_matplotlib_cleared')
 def clear_matplotlib():
+    if matplotlib is None:
+        pytest.skip('Matplotlib is not installed.')
     for should_yield in (True, False):
         to_delete = [module_name
                      for module_name in sys.modules
@@ -21,16 +24,7 @@ def clear_matplotlib():
             yield True
 
 
-def replace_image(report):
-    report = trim_report(report)
-    report = re.sub(r"image='[a-zA-Z0-9+/=]*'", "image='...'", report)
-    return report
-
-
-@pytest.mark.skipif(sys.version_info >= MATPLOTLIB_UNSUPPORTED_VERSION,
-                    reason='matplotlib not yet supported.')
 def test_show(is_matplotlib_cleared):
-    assert is_matplotlib_cleared
     code = """\
 import matplotlib.pyplot as plt
 
@@ -51,6 +45,27 @@ create_image
     assert expected_report == replace_image(report)
 
 
+def test_show_with_canvas_size(is_matplotlib_cleared):
+    code = """\
+import matplotlib.pyplot as plt
+
+data = [1, 2]
+plt.plot(data)
+plt.show()
+"""
+    expected_report = """\
+create_image
+    0
+    0
+    image='...'
+"""
+    tracer = TraceRunner()
+
+    report = tracer.trace_turtle(code, width=400, height=200)
+
+    assert expected_report == replace_image(report)
+
+
 def test_numpy_random():
     code = """\
 import numpy as np 
